# Copyright (c) 2016-2017, Jani Nikula <jani@nikula.org>
# Licensed under the terms of BSD 2-Clause, see LICENSE for details.
"""
Hawkmoth
========

Sphinx C Domain autodoc directive extension.
"""

import glob
import os
from typing import Optional

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.statemachine import ViewList
from sphinx import addnodes
from sphinx.util.nodes import nested_parse_with_titles
from sphinx.util.docutils import switch_source_input, SphinxDirective
from sphinx.util import logging

from hawkmoth.util import compiler
from hawkmoth.parser import parse, ErrorLevel
from hawkmoth.util import strutil
from hawkmoth import docstring

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)),
                       'VERSION')) as version_file:
    __version__ = version_file.read().strip()

class _AutoBaseDirective(SphinxDirective):
    logger = logging.getLogger(__name__)

    option_spec = {
        'transform': directives.unchanged_required,
        'clang': strutil.string_list,
    }
    has_content = False

    _domain: Optional[str] = None
    _docstring_types: Optional[list[type[docstring.Docstring]]] = None

    def __display_parser_diagnostics(self, errors):
        # Map parser diagnostic level to Sphinx level name
        log_level_map = {
            ErrorLevel.DEBUG: 'DEBUG',
            ErrorLevel.INFO: 'VERBOSE',
            ErrorLevel.WARNING: 'WARNING',
            ErrorLevel.ERROR: 'ERROR',
            ErrorLevel.CRITICAL: 'ERROR',
        }

        for error in errors:
            self.logger.log(log_level_map[error.level], error.get_message(),
                            location=(self.env.docname, self.lineno))

    def __get_clang_args(self):
        clang_args = self.env.config.hawkmoth_clang.copy()

        if self._domain == 'c':
            clang_args.extend(self.env.config.hawkmoth_clang_c.copy())
            clang_args.extend(self.options.get('clang', []))
            clang_args.extend(self.env.config._clang_args_post_c.copy())
        else:
            clang_args.extend(self.env.config.hawkmoth_clang_cpp.copy())
            clang_args.extend(self.options.get('clang', []))
            clang_args.extend(self.env.config._clang_args_post_cpp.copy())

        return clang_args

    def __parse(self, filename):
        clang_args = self.__get_clang_args()

        # Cached parse results per rst document
        parsed_files = self.env.temp_data.setdefault('hawkmoth_parsed_files', {})

        # The output depends on domain and clang args
        key = (filename, self._domain, tuple(clang_args))

        if key in parsed_files:
            return

        # Tell Sphinx about the dependency
        self.env.note_dependency(filename)

        docstrings, errors = parse(filename, domain=self._domain,
                                   clang_args=clang_args)

        self.__display_parser_diagnostics(errors)

        parsed_files[key] = docstrings

    def __parsed_files(self, filter_filenames=None, filter_domains=None,
                       filter_clang_args=None):
        parsed_files = self.env.temp_data.get('hawkmoth_parsed_files', {})

        for root in parsed_files.values():
            if filter_filenames is not None and root.get_filename() not in filter_filenames:
                continue

            if filter_domains is not None and root.get_domain() not in filter_domains:
                continue

            if filter_clang_args is not None and root.get_clang_args() not in filter_clang_args:
                continue

            yield root

    def __process_docstring(self, lines):
        transform = self.options.get('transform', self.env.config.hawkmoth_transform_default)

        self.env.app.emit('hawkmoth-process-docstring', lines, transform, self.options)

    def __get_docstrings_for_root(self, viewlist, root):
        def process_docstring(lines): self.__process_docstring(lines)

        num_matches = 0
        for docstrings in root.walk(recurse=False, filter_types=self._docstring_types,
                                    filter_names=self._get_names()):
            num_matches += 1
            for docstr in docstrings.walk(filter_names=self._get_members()):
                lines, line_number = docstr.get_docstring(process_docstring=process_docstring)
                for line in lines:
                    # viewlist line numbers are 0-based
                    viewlist.append(line, root.get_filename(), line_number - 1)
                    line_number += 1

        return num_matches

    def __get_docstrings(self, viewlist):
        num_matches = 0
        for root in self.__parsed_files(filter_filenames=self._get_filenames(),
                                        filter_domains=[self._domain],
                                        filter_clang_args=[self.__get_clang_args()]):
            num_matches += self.__get_docstrings_for_root(viewlist, root)

        if num_matches == 0:
            if self._get_names():
                args = ' '.join(self.arguments)
                self.logger.warning(f'"{self.name}:: {args}" does not match documented symbols.',
                                    location=(self.env.docname, self.lineno))
            else:
                # autodoc
                self.logger.warning('No documented symbols were found.',
                                    location=(self.env.docname, self.lineno))
        elif num_matches > 1 and self._get_names():
            args = ' '.join(self.arguments)
            self.logger.warning(f'"{self.name}:: {args}" matches {num_matches} documented symbols.',
                                location=(self.env.docname, self.lineno))

    def _get_names(self):
        return None

    def _get_members(self):
        return None

    def _get_filenames(self):
        raise NotImplementedError(self.__class__.__name__ + '._get_filenames')

    def run(self):
        if self._get_filenames():
            for filename in self._get_filenames():
                self.__parse(filename)

        result = ViewList()

        self.__get_docstrings(result)

        # Parse the extracted reST
        with switch_source_input(self.state, result):
            node = nodes.section()
            nested_parse_with_titles(self.state, result, node)

        return node.children

class _AutoDocDirective(_AutoBaseDirective):
    """Extract all documentation comments from the specified files"""

    # Allow passing a variable number of file patterns as arguments
    required_arguments = 1
    optional_arguments = 100   # arbitrary limit

    def _get_filenames(self):
        ret = []
        for pattern in self.arguments:
            filenames = glob.glob(os.path.join(self.env.config.hawkmoth_root, pattern))
            if len(filenames) == 0:
                self.logger.warning(f'Pattern "{pattern}" does not match any files.',
                                    location=(self.env.docname, self.lineno))
                continue

            for filename in filenames:
                if os.path.isfile(filename):
                    ret.append(os.path.abspath(filename))
                else:
                    self.logger.warning(f'Path "{filename}" matching pattern "{pattern}" is not a file.',  # noqa: E501
                                        location=(self.env.docname, self.lineno))

        return ret

# Base class for named stuff
class _AutoSymbolDirective(_AutoBaseDirective):
    """Extract specified documentation comments from the specified file"""

    required_arguments = 1
    optional_arguments = 0

    option_spec = _AutoBaseDirective.option_spec.copy()
    option_spec.update({
        'file': directives.unchanged_required,
    })

    def _get_filenames(self):
        filename = self.options.get('file')
        if not filename:
            return None

        return [os.path.abspath(os.path.join(self.env.config.hawkmoth_root, filename))]

    def _get_names(self):
        return [self.arguments[0]]

def _members_filter(argument):
    # Use None for members option without an argument to not filter.
    if argument is None:
        return None
    return strutil.string_list(argument)

class _AutoCompoundDirective(_AutoSymbolDirective):
    option_spec = _AutoSymbolDirective.option_spec.copy()
    option_spec.update({
        'members': _members_filter,
    })

    def _get_members(self):
        # By default use [] as a filter that does not match any members.
        return self.options.get('members', [])

class CAutoDocDirective(_AutoDocDirective):
    _domain = 'c'

class CAutoSectionDirective(_AutoSymbolDirective):
    # Allow spaces in the directive argument (the name)
    final_argument_whitespace = True
    _domain = 'c'
    _docstring_types = [docstring.TextDocstring]

class CAutoVarDirective(_AutoSymbolDirective):
    _domain = 'c'
    _docstring_types = [docstring.VarDocstring]

class CAutoTypeDirective(_AutoSymbolDirective):
    _domain = 'c'
    _docstring_types = [docstring.TypedefDocstring]

class CAutoMacroDirective(_AutoSymbolDirective):
    _domain = 'c'
    _docstring_types = [docstring.MacroDocstring, docstring.MacroFunctionDocstring]

class CAutoFunctionDirective(_AutoSymbolDirective):
    _domain = 'c'
    _docstring_types = [docstring.FunctionDocstring]

class CAutoStructDirective(_AutoCompoundDirective):
    _domain = 'c'
    _docstring_types = [docstring.StructDocstring]

class CAutoUnionDirective(_AutoCompoundDirective):
    _domain = 'c'
    _docstring_types = [docstring.UnionDocstring]

class CAutoEnumDirective(_AutoCompoundDirective):
    _domain = 'c'
    _docstring_types = [docstring.EnumDocstring]

class CppAutoDocDirective(_AutoDocDirective):
    _domain = 'cpp'

class CppAutoSectionDirective(_AutoSymbolDirective):
    # Allow spaces in the directive argument (the name)
    final_argument_whitespace = True
    _domain = 'cpp'
    _docstring_types = [docstring.TextDocstring]

class CppAutoVarDirective(_AutoSymbolDirective):
    _domain = 'cpp'
    _docstring_types = [docstring.VarDocstring]

class CppAutoTypeDirective(_AutoSymbolDirective):
    _domain = 'cpp'
    _docstring_types = [docstring.TypedefDocstring, docstring.TypeAliasDocstring]

class CppAutoMacroDirective(_AutoSymbolDirective):
    _domain = 'cpp'
    _docstring_types = [docstring.MacroDocstring, docstring.MacroFunctionDocstring]

class CppAutoFunctionDirective(_AutoSymbolDirective):
    _domain = 'cpp'
    _docstring_types = [docstring.FunctionDocstring]

class CppAutoStructDirective(_AutoCompoundDirective):
    _domain = 'cpp'
    _docstring_types = [docstring.StructDocstring]

class CppAutoUnionDirective(_AutoCompoundDirective):
    _domain = 'cpp'
    _docstring_types = [docstring.UnionDocstring]

class CppAutoEnumDirective(_AutoCompoundDirective):
    _domain = 'cpp'
    _docstring_types = [docstring.EnumDocstring]

class CppAutoClassDirective(_AutoCompoundDirective):
    _domain = 'cpp'
    _docstring_types = [docstring.ClassDocstring]

def _uri_format(env, signode):
    """Generate a source URI for signode"""
    uri_template = env.config.hawkmoth_source_uri

    if signode.source is None or signode.line is None:
        return None

    source = os.path.relpath(signode.source, start=env.config.hawkmoth_root)

    # Note: magic +1 to take into account we've added signode ourselves and its
    # not present in source
    uri = uri_template.format(source=source, line=signode.line + 1)

    return uri

def _doctree_read(app, doctree):
    env = app.builder.env

    # Bail out early if not configured
    uri_template = env.config.hawkmoth_source_uri
    if uri_template is None:
        return

    for objnode in list(doctree.findall(addnodes.desc)):
        if objnode.get('domain') not in ['c', 'cpp']:
            continue

        for signode in objnode:
            if not isinstance(signode, addnodes.desc_signature):
                continue

            uri = _uri_format(env, signode)
            if not uri:
                continue

            # Similar to sphinx.ext.linkcode
            inline = nodes.inline('', '[source]', classes=['viewcode-link'])
            onlynode = addnodes.only(expr='html')
            onlynode += nodes.reference('', '', inline, internal=False, refuri=uri)
            signode += onlynode

def _autoconf(app, config):
    logger = logging.getLogger(__name__)
    cpath = config.hawkmoth_compiler
    autoconf = config.hawkmoth_autoconf if config.hawkmoth_autoconf else []

    ignored_options = [x for x in autoconf if x not in ['stdinc']]
    if len(ignored_options) > 0:
        logger.warning(f'autoconf: {ignored_options} unsupported option(s) ignored')

    config._clang_args_post_c = []
    config._clang_args_post_cpp = []

    if 'stdinc' in autoconf:
        if cpath:
            config._clang_args_post_c = compiler.get_include_args(cpath, 'c')
            config._clang_args_post_cpp = compiler.get_include_args(cpath, 'c++')
        else:
            logger.warning('autoconf: \'stdinc\' option ignored (missing compiler)')

    logger.verbose(f'autoconf: Using C include args: {config._clang_args_post_c}')
    logger.verbose(f'autoconf: Using C++ include args: {config._clang_args_post_cpp}')

def setup(app):
    app.require_sphinx('3.0')

    app.add_config_value('hawkmoth_root', app.confdir, 'env', [str])
    app.add_config_value('hawkmoth_compiler', 'clang', 'env', [str, type(None)])
    app.add_config_value('hawkmoth_autoconf', ['stdinc'], 'env', [list[str], type(None)])
    app.add_config_value('hawkmoth_clang', [], 'env', [list[str]])
    app.add_config_value('hawkmoth_clang_c', [], 'env', [list[str]])
    app.add_config_value('hawkmoth_clang_cpp', [], 'env', [list[str]])

    app.add_config_value('hawkmoth_transform_default', None, 'env', [str, type(None)])

    app.add_directive_to_domain('c', 'autodoc', CAutoDocDirective)
    app.add_directive_to_domain('c', 'autosection', CAutoSectionDirective)
    app.add_directive_to_domain('c', 'autovar', CAutoVarDirective)
    app.add_directive_to_domain('c', 'autotype', CAutoTypeDirective)
    app.add_directive_to_domain('c', 'autostruct', CAutoStructDirective)
    app.add_directive_to_domain('c', 'autounion', CAutoUnionDirective)
    app.add_directive_to_domain('c', 'autoenum', CAutoEnumDirective)
    app.add_directive_to_domain('c', 'automacro', CAutoMacroDirective)
    app.add_directive_to_domain('c', 'autofunction', CAutoFunctionDirective)

    app.add_directive_to_domain('cpp', 'autodoc', CppAutoDocDirective)
    app.add_directive_to_domain('cpp', 'autosection', CppAutoSectionDirective)
    app.add_directive_to_domain('cpp', 'autovar', CppAutoVarDirective)
    app.add_directive_to_domain('cpp', 'autotype', CppAutoTypeDirective)
    app.add_directive_to_domain('cpp', 'autostruct', CppAutoStructDirective)
    app.add_directive_to_domain('cpp', 'autounion', CppAutoUnionDirective)
    app.add_directive_to_domain('cpp', 'autoenum', CppAutoEnumDirective)
    app.add_directive_to_domain('cpp', 'automacro', CppAutoMacroDirective)
    app.add_directive_to_domain('cpp', 'autofunction', CppAutoFunctionDirective)
    app.add_directive_to_domain('cpp', 'autoclass', CppAutoClassDirective)

    app.add_event('hawkmoth-process-docstring')

    # Auto configure once during initialization, after Sphinx config type checks
    app.connect('config-inited', _autoconf, priority=850)

    # Source code link
    app.add_config_value('hawkmoth_source_uri', None, 'env', [str, type(None)])
    app.connect('doctree-read', _doctree_read)

    return dict(version=__version__,
                parallel_read_safe=True, parallel_write_safe=True)



# Copyright (c) 2016-2021 Jani Nikula <jani@nikula.org>
# Copyright (c) 2018-2023 Bruno Santos <brunomanuelsantos@tecnico.ulisboa.pt>
# Licensed under the terms of BSD 2-Clause, see LICENSE for details.
"""
Documentation comment storage and converter
===========================================

Class hierarchy for storing a tree of different types of documentation
comments.

The documentation comments from C source are stored verbatim. There are methods
for walking the documentation comment tree and returning the documentation with
conversions:

* Stripping the comment delimiters (``/**`` and ``*/``) and continuation
  line prefixes (e.g. ``␣*␣``).

* Optional transformation of the documentation comments to support different
  syntaxes. These filters are expected to translate the comment into
  reStructuredText.

* Generation of Sphinx C Domain directives with appropriate indentation.
"""

import hashlib
import os
import re

from docutils import statemachine

def _commonprefix_len(lines):
    # common prefix
    prefix = os.path.commonprefix(lines)

    # common prefix length of limited characters
    return len(prefix) - len(prefix.lstrip(' \t*'))

def _get_prefix_len(lines):
    # ignore lines with just space
    lines = [line for line in lines if line.strip()]
    prefix_len = _commonprefix_len(lines)

    # ignore lines with just the prefix and space
    lines = [line for line in lines if line[prefix_len:].strip()]
    prefix_len = _commonprefix_len(lines)

    return prefix_len

class Docstring():
    _indent = 0
    _fmt = ''

    def __init__(self, cursor=None, text=None, meta=None, nest=0):
        if cursor:
            self._args = cursor.args
            self._decl_name = cursor.decl_name
            self._domain = cursor.domain
            self._meta = cursor.meta
            self._name = cursor.name
            self._quals = cursor.quals
            self._text = cursor.comment
            self._ttype = cursor.type
        else:
            self._args = None
            self._decl_name = None
            self._domain = None
            self._meta = meta
            self._name = None
            self._quals = None
            self._text = text
            self._ttype = None

        self._nest = nest
        self._children = []

    def _match(self, filter_types=None, filter_names=None):
        if filter_types is not None and type(self) not in filter_types:
            return False

        if filter_names is not None and self.get_name() not in filter_names:
            return False

        return True

    def walk(self, recurse=True, filter_types=None, filter_names=None):
        if self._text:
            yield self

    @staticmethod
    def is_doc(comment):
        """Test if comment is a C documentation comment."""
        return comment.startswith('/**') and comment != '/**/'

    def _get_header_lines(self):
        name = self._get_decl_name()
        domain = self._domain

        header = self._fmt.format(domain=domain, name=name)

        return header.splitlines()

    def _get_comment_lines(self):
        return statemachine.string2lines(self._text, 8, convert_whitespace=True)

    @staticmethod
    def _remove_comment_markers(lines):
        """Remove comment markers and line prefixes from comment lines.

        Return the number of lines removed from the beginning.
        """
        line_offset = 0

        lines[0] = re.sub(r'^/\*\*[ \t]*', '', lines[0])
        lines[-1] = re.sub(r'[ \t]*\*/$', '', lines[-1])

        prefix_len = _get_prefix_len(lines[1:-1])
        lines[1:-1] = [line[prefix_len:] for line in lines[1:-1]]

        while lines and (not lines[0] or lines[0].isspace()):
            line_offset += 1
            del lines[0]

        while lines and (not lines[-1] or lines[-1].isspace()):
            del lines[-1]

        return line_offset

    @staticmethod
    def _nest_lines(lines, nest):
        """
        Indent documentation block for nesting.

        Args:
            lines (List[str]): Documentation body as list of strings,
                modified in-place.
            nest (int): Nesting level. For each level, the final block is indented
                one level. Useful for (e.g.) declaring structure members.
        """
        lines[:] = ['   ' * nest + line if line else '' for line in lines]

    def get_docstring(self, process_docstring=None):
        header_lines = self._get_header_lines()
        comment_lines = self._get_comment_lines()

        line_offset = Docstring._remove_comment_markers(comment_lines)

        if process_docstring is not None:
            process_docstring(comment_lines)

        Docstring._nest_lines(comment_lines, self._indent)

        # ensure we have cushion blank line before the docstring
        if len(header_lines) == 0 or header_lines[0] != '':
            header_lines.insert(0, '')

        # ensure we have cushion blank line between header and comment
        if header_lines[-1] != '':
            header_lines.append('')

        line_offset -= len(header_lines)

        lines = header_lines + comment_lines

        Docstring._nest_lines(lines, self._nest)

        # ensure we have cushion blank line after the docstring
        if lines[-1] != '':
            lines.append('')

        return lines, self.get_line() + line_offset

    def get_meta(self):
        return self._meta

    def _get_decl_name(self):
        return self._decl_name if self._decl_name else self._name

    def get_name(self):
        return self._name

    def get_line(self):
        return self._meta['line']

class TextDocstring(Docstring):
    _indent = 0
    _fmt = ''

    def get_name(self):
        """Figure out a name for the text comment based on the comment contents.

        The name is the sub-string starting from the first alphanumeric
        character in the comment to the next :, ., or newline.

        This sensibly covers cases like reStructuredText hyperlink targets::

            .. _Foo Bar:

        and section titles::

            Foo Bar
            =======

        and just first sentences::

            Foo Bar. Blah.

        Not perfect, but good enough.
        """
        # If the parser passed in a name, use it (unlikely)
        if self._name:
            return self._name

        mo = re.search(r'[\W_]*(?P<name>\w[^:.\n\r]*)', self._text)

        return mo.group('name') if mo else None

class VarDocstring(Docstring):
    _indent = 1
    _fmt = '.. {domain}:var:: {ttype}{type_spacer}{name}'

    def _get_header_lines(self):
        name = self._get_decl_name()
        domain = self._domain
        ttype = self._ttype

        type_spacer = ''
        if ttype and not (len(ttype) == 0 or ttype.endswith('*')):
            type_spacer = ' '

        header = self._fmt.format(domain=domain, name=name, ttype=ttype,
                                  type_spacer=type_spacer)

        return header.splitlines()

class TypedefDocstring(Docstring):
    _indent = 1
    _fmt = '.. {domain}:type:: {name}'

class TypeAliasDocstring(Docstring):
    _indent = 1
    _fmt = '.. cpp:type:: {name} = {underlying_type}'

    def __init__(self, cursor, nest):
        self._underlying_type = cursor.value
        super().__init__(cursor=cursor, nest=nest)

    def _get_header_lines(self):
        name = self._get_decl_name()
        underlying_type = self._underlying_type.spelling

        header = self._fmt.format(name=name, underlying_type=underlying_type)

        return header.splitlines()

class _CompoundDocstring(Docstring):
    def _get_decl_name(self):
        # If decl_name is empty, it means this is an anonymous declaration.
        if self._decl_name is None:
            # Sphinx expects @name for anonymous entities. The name must be both
            # stable and unique. Create one.
            decl_name = hashlib.md5(f'{self._text}{self.get_line()}'.encode()).hexdigest()

            return f'@anonymous_{decl_name}'

        return self._decl_name

    def add_child(self, comment):
        self._children.append(comment)

    def add_children(self, comments):
        self._children.extend(comments)

    def walk(self, recurse=True, filter_types=None, filter_names=None):
        # Note: The filtering is pretty specialized for our use case here. It
        # only filters the immediate children, not this comment, nor
        # grandchildren.

        # The contents of the parent will always be before children.
        if self._text:
            yield self

        # Sort the children by order of appearance. We may add other sort
        # options later.
        for comment in sorted(self._children, key=lambda c: c.get_line()):
            if comment._match(filter_types=filter_types, filter_names=filter_names):
                if recurse:
                    yield from comment.walk()
                else:
                    yield comment

class RootDocstring(_CompoundDocstring):
    def __init__(self, filename=None, domain='c', clang_args=None):
        super().__init__()
        self._filename = filename
        self._domain = domain
        self._clang_args = clang_args

    def get_filename(self):
        return self._filename

    def get_clang_args(self):
        return self._clang_args

    def get_domain(self):
        return self._domain

class StructDocstring(_CompoundDocstring):
    _indent = 1
    _fmt = '.. {domain}:struct:: {name}'

class UnionDocstring(_CompoundDocstring):
    _indent = 1
    _fmt = '.. {domain}:union:: {name}'

class EnumDocstring(_CompoundDocstring):
    _indent = 1
    _fmt = '.. {domain}:enum:: {name}'

class EnumeratorDocstring(Docstring):
    _indent = 1
    _fmt = '.. {domain}:enumerator:: {name}{value}'

    def __init__(self, cursor, nest=0):
        self._value = cursor.value
        super().__init__(cursor=cursor, nest=nest)

    def _get_header_lines(self):
        value = f' = {self._value}' if self._value is not None else ''
        header = self._fmt.format(domain=self._domain, name=self._get_decl_name(),
                                  value=value)

        return header.splitlines()

class MemberDocstring(Docstring):
    _indent = 1
    _fmt = '.. {domain}:member:: {ttype}{type_spacer}{name}'

    def _get_header_lines(self):
        name = self._get_decl_name()
        domain = self._domain
        ttype = self._ttype

        type_spacer = ''
        if ttype and not (len(ttype) == 0 or ttype.endswith('*')):
            type_spacer = ' '

        header = self._fmt.format(domain=domain, name=name, ttype=ttype,
                                  type_spacer=type_spacer)

        return header.splitlines()

class MacroDocstring(Docstring):
    _indent = 1
    _fmt = '.. c:macro:: {name}'

class MacroFunctionDocstring(Docstring):
    _indent = 1
    _fmt = '.. c:macro:: {name}({args})'

    def _get_header_lines(self):
        name = self._get_decl_name()
        args = ', '.join([n for _, n in self._args])

        header = self._fmt.format(name=name, args=args)

        return header.splitlines()

class FunctionDocstring(Docstring):
    _indent = 1
    _fmt = '.. {domain}:function:: {ttype}{type_spacer}{name}({args}){quals_spacer}{quals}'

    def _get_header_lines(self):
        name = self._get_decl_name()
        domain = self._domain
        ttype = self._ttype
        quals = self._quals

        type_spacer = ''
        if ttype and not (len(ttype) == 0 or ttype.endswith('*')):
            type_spacer = ' '

        quals_spacer = ''
        if quals and len(quals) > 0:
            quals_spacer = ' '

        args = ''
        if self._args and len(self._args) > 0:
            def pad_type(t): return '' if len(t) == 0 or t.endswith('*') or t.endswith('&') else ' '
            def arg_fmt(t, n): return f'{t}{pad_type(t)}{n}'
            args = ', '.join([arg_fmt(t, n) for t, n in self._args])

        header = self._fmt.format(domain=domain, name=name, ttype=ttype,
                                  type_spacer=type_spacer, args=args,
                                  quals=quals, quals_spacer=quals_spacer)

        return header.splitlines()

class ClassDocstring(_CompoundDocstring):
    _indent = 1
    _fmt = '.. cpp:class:: {name}'

class EnumClassDocstring(_CompoundDocstring):
    _indent = 1
    _fmt = '.. cpp:enum-class:: {name}'



# Copyright (c) 2023, Jani Nikula <jani@nikula.org>
# Licensed under the terms of BSD 2-Clause, see LICENSE for details.

import re
from typing import Optional

# The "operator" character, either \ or @, but not escaped with \
OP = r'(?<!\\)(?P<op>[\\@])'

class _handler:
    """Base class for all command handlers."""
    _indented_paragraph = False

    def __init__(self, app=None, indent=None, op=None, command=None, rest=None):
        self._app = app
        self._indent = indent
        self._op = op
        self._command = command
        self._rest = rest

    def blank_line_ends(self):
        """Does a blank line end this command?"""
        return True

    def command_ends(self, command):
        """Does the provided command end this command?"""
        return True

    @staticmethod
    def _inline_markup(line):
        """Handle inline markup."""

        word_regex = r'[^\s.]+'
        tagged_phrase_regex = r'[^<]*'

        # italics: \a \e \em <em>...</em>
        line = re.sub(fr'{OP}(a|e|em)\s+(?P<markup>{word_regex})', r'*\g<markup>*', line)
        line = re.sub(fr'<em>(?P<markup>{tagged_phrase_regex})</em>', r'*\g<markup>*', line)

        # bold: \b <b>...</b>
        line = re.sub(fr'{OP}b\s+(?P<markup>{word_regex})', r'**\g<markup>**', line)
        line = re.sub(fr'<b>(?P<markup>{tagged_phrase_regex})</b>', r'**\g<markup>**', line)

        # monospace: \c \p <tt>...</tt>
        line = re.sub(fr'{OP}(c|p)\s+(?P<markup>{word_regex})', r'``\g<markup>``', line)
        line = re.sub(fr'<tt>(?P<markup>{tagged_phrase_regex})</tt>', r'``\g<markup>``', line)

        # references to previous anchors
        # FIXME: link title
        line = re.sub(fr'{OP}ref\s+(?P<ref>\w+)', r':ref:`\g<ref>`', line)

        # FIXME:
        # - copybrief
        # - copydetails
        # - emoji
        # - f
        # - n

        return line

    def rest(self):
        """Return the "rest" of the line after @command."""
        return self._inline_markup(self._rest)

    def header(self):
        """Convert the first line of this command."""
        yield ''

    def convert(self, line):
        """Convert a regular line within this paragraph."""
        line = self._inline_markup(line)

        if self._indented_paragraph:
            line = f'   {line}'

        yield line

class _plain(_handler):
    pass

class _not_implemented(_plain):
    """Placeholder for commands that have not been implemented."""
    # FIXME: warn about not implemented commands
    pass

class _block_with_end_command(_handler):
    """Paragraph with a dedicated command to end it.

    For example, @code/@endcode."""
    _end_command: Optional[str] = None

    def end_command(self):
        """Get the name of the command that ends this paragraph."""
        return self._end_command if self._end_command else f'end{self._command}'

    def blank_line_ends(self):
        return False

    def command_ends(self, command):
        return self.end_command() == command

class _ignore_until_end_command(_block_with_end_command):
    """Ignore the paragraph until dedicated command ends it."""
    # FIXME: warn about ignored commands
    def header(self):
        yield ''

    def convert(self, line):
        yield ''

class _startuml(_ignore_until_end_command):
    # Needed because it's @startuml/@enduml, not @uml/@enduml.
    _end_command = 'enduml'

class _code(_block_with_end_command):
    def header(self):
        yield ''
        yield '.. code-block::'
        yield ''

    def convert(self, line):
        yield f'   {line}'

class _anchor(_handler):
    def header(self):
        anchor = self._rest.strip()

        yield ''
        yield f'.. {anchor}:'
        yield ''

class _strip_command(_handler):
    """Strip command, treat everything else as normal."""
    def header(self):
        yield f'{self._indent}{self.rest().strip()}'

class _field_list(_handler):
    """Paragraph which becomes a single field list item."""
    _field_name: Optional[str] = None
    _indented_paragraph = True

    def field_name(self):
        return self._field_name if self._field_name else self._command

    def header(self):
        yield ''
        yield f'{self._indent}:{self.field_name()}:{self.rest()}'

class _author(_field_list):
    _field_name = 'author'

class _return(_field_list):
    _field_name = 'return'

class _see(_field_list):
    _field_name = 'see'

class _param(_field_list):
    """Parameter description."""
    _field_name = 'param'

    def header(self):
        mo = re.match(r'^((?P<sp0>\s*)\[(?P<direction>[a-zA-Z, ]+)\])?(?P<sp1>\s*)(?P<name>([a-zA-Z0-9_]+|\.\.\.))(?P<sp2>\s*(?P<desc>.*))',  # noqa: E501
                      self.rest())
        if mo is None:
            # FIXME
            yield ''
            return

        direction = mo.group('direction')
        name = mo.group('name')
        desc = mo.group('desc')

        yield ''
        if direction:
            yield f'{self._indent}:param {name}: **[{direction}]** {desc}'
        else:
            yield f'{self._indent}:param {name}: {desc}'

class _admonition(_handler):
    """Admonitions such as @note and @warning."""
    _indented_paragraph = True
    _directive = None

    def directive(self):
        return self._directive if self._directive else self._command

    def header(self):
        yield ''
        yield f'.. {self.directive()}::'
        yield ''
        rest = self.rest().strip()
        if rest:
            yield f'   {rest}'

# Map non-inline commands to handler classes.
#
# All the non-inline commands in the order listed at
# https://www.doxygen.nl/manual/commands.html
_handlers = {
    # structural indicators
    'addtogroup': _not_implemented,
    'callgraph': _not_implemented,
    'hidecallgraph': _not_implemented,
    'callergraph': _not_implemented,
    'hidecallergraph': _not_implemented,
    'showrefby': _not_implemented,
    'hiderefby': _not_implemented,
    'showrefs': _not_implemented,
    'hiderefs': _not_implemented,
    'showinlinesource': _not_implemented,
    'hideinlinesource': _not_implemented,
    'includegraph': _not_implemented,
    'hideincludegraph': _not_implemented,
    'includedbygraph': _not_implemented,
    'hideincludedbygraph': _not_implemented,
    'directorygraph': _not_implemented,
    'hidedirectorygraph': _not_implemented,
    'collaborationgraph': _not_implemented,
    'hidecollaborationgraph': _not_implemented,
    'inheritancegraph': _not_implemented,
    'hideingeritancegraph': _not_implemented,
    'groupgraph': _not_implemented,
    'hidegroupgraph': _not_implemented,
    'qualifier': _not_implemented,
    'category': _not_implemented,
    'class': _not_implemented,  # WARN
    'concept': _not_implemented,
    'def': _not_implemented,  # WARN
    'defgroup': _not_implemented,
    'dir': _not_implemented,
    'enum': _not_implemented,  # WARN
    'example': _not_implemented,  # FIXME
    'endinternal': _not_implemented,  # WARN
    'extends': _not_implemented,  # FIXME
    'file': _not_implemented,  # FIXME
    'fileinfo': _not_implemented,
    'lineinfo': _not_implemented,  # WARN
    'fn': _not_implemented,  # WARN
    'headerfile': _not_implemented,
    'hideinitializer': _not_implemented,
    'idlexcept': _not_implemented,
    'implements': _not_implemented,
    'ingroup': _not_implemented,
    'interface': _not_implemented,  # WARN
    'internal': _not_implemented,
    'mainpage': _not_implemented,  # FIXME
    'memberof': _not_implemented,
    'module': _not_implemented,
    'name': _not_implemented,
    'namespace': _not_implemented,
    'nosubgrouping': _not_implemented,
    'overload': _not_implemented,
    'package': _not_implemented,
    'page': _not_implemented,  # FIXME
    'private': _not_implemented,
    'privatesection': _not_implemented,
    'property': _not_implemented,
    'protected': _not_implemented,
    'protectedsection': _not_implemented,
    'protocol': _not_implemented,
    'public': _not_implemented,
    'publicsection': _not_implemented,
    'pure': _not_implemented,
    'relates': _not_implemented,  # FIXME
    'related': _not_implemented,  # FIXME
    'relatesalso': _not_implemented,
    'relatedalso': _not_implemented,
    'showinitializer': _not_implemented,
    'static': _not_implemented,
    'typedef': _not_implemented,  # WARN
    'union': _not_implemented,  # WARN
    'var': _not_implemented,  # WARN
    'vhdlflow': _not_implemented,
    'weakgroup': _not_implemented,
    # section indicators
    'attention': _admonition,
    'author': _author,
    'authors': _author,
    'brief': _strip_command,
    'bug': _field_list,
    'cond': _not_implemented,
    'copyright': _field_list,
    'date': _field_list,
    'showdate': _not_implemented,
    'deprecated': _field_list,
    'details': _strip_command,
    'noop': _not_implemented,
    'raisewarning': _not_implemented,
    'else': _not_implemented,
    'elseif': _not_implemented,
    'endcond': _not_implemented,
    'endif': _not_implemented,
    'exception': _field_list,
    'if': _not_implemented,
    'ifnot': _not_implemented,
    'invariant': _field_list,
    'note': _admonition,
    'par': _not_implemented,
    'param': _param,
    'parblock': _not_implemented,
    'endparblock': _not_implemented,
    'tparam': _field_list,
    'post': _field_list,
    'pre': _field_list,
    'remark': _field_list,
    'remarks': _field_list,
    'result': _return,
    'return': _return,
    'returns': _return,
    'retval': _return,
    'sa': _see,
    'see': _see,
    'short': _strip_command,
    'since': _field_list,
    'test': _field_list,
    'throw': _field_list,
    'throws': _field_list,
    'todo': _field_list,
    'version': _field_list,
    'warning': _admonition,
    'xrefitem': _not_implemented,
    'addindex': _not_implemented,
    'anchor': _anchor,
    'cite': _field_list,
    'endlink': _not_implemented,
    'link': _not_implemented,
    'refitem': _not_implemented,
    'secreflist': _not_implemented,
    'endsecreflist': _not_implemented,
    'subpage': _not_implemented,
    'tableofcontents': _not_implemented,
    'section': _not_implemented,
    'subsection': _not_implemented,
    'subsubsection': _not_implemented,
    'paragraph': _not_implemented,
    'dontinclude': _not_implemented,
    'include': _not_implemented,
    'includelineno': _not_implemented,
    'includedoc': _not_implemented,
    'line': _not_implemented,
    'skip': _not_implemented,
    'skipline': _not_implemented,
    'snippet': _not_implemented,
    'snippetlineno': _not_implemented,
    'snippetdoc': _not_implemented,
    'until': _not_implemented,
    'verbinclude': _not_implemented,
    'htmlinclude': _not_implemented,
    'latexinclude': _not_implemented,
    'rtfinclude': _not_implemented,
    'maninclude': _not_implemented,
    'docbookinclude': _not_implemented,
    'xmlinclude': _not_implemented,
    # visual enhancements
    'arg': _not_implemented,  # FIXME
    'code': _code,
    'copydoc': _not_implemented,
    'docbookonly': _ignore_until_end_command,
    'dot': _ignore_until_end_command,
    'msc': _ignore_until_end_command,
    'startuml': _startuml,
    'dotfile': _not_implemented,
    'mscfile': _not_implemented,
    'diafile': _not_implemented,
    'doxyconfig': _not_implemented,
    'endcode': _plain,
    'enddocbookonly': _plain,
    'enddot': _plain,
    'endmsc': _plain,
    'enduml': _plain,
    'endhtmlonly': _plain,
    'endlatexonly': _plain,
    'endmanonly': _plain,
    'endrtfonly': _plain,
    'endverbatim': _plain,
    'endxmlonly': _plain,
    'htmlonly': _ignore_until_end_command,
    'image': _not_implemented,  # FIXME
    'latexonly': _ignore_until_end_command,
    'manonly': _ignore_until_end_command,
    'li': _not_implemented,  # FIXME
    'rtfonly': _ignore_until_end_command,
    'verbatim': _code,
    'xmlonly': _ignore_until_end_command,
}

# Ensure at least this regex is compiled.
_command_pattern = re.compile(fr'(?P<indent>\s*){OP}(?P<command>[a-zA-Z]+)(?P<rest>.*)')

def _convert(lines, app=None):
    handler = _plain(app=app)

    for line in lines:
        if line.strip() == '' and handler.blank_line_ends():
            handler = _plain(app=app)
            yield from handler.convert(line)
            continue

        mo = _command_pattern.match(line)
        if mo is None:
            # No command match, continue with current handler
            yield from handler.convert(line)
            continue

        command = mo.group('command')

        handler_cls = _handlers.get(command)
        if handler_cls is None:
            # Unknown command, continue with current handler
            yield from handler.convert(line)
            continue

        if not handler.command_ends(command):
            # Command does not finish block, continue with current handler
            yield from handler.convert(line)
            continue

        # Switch paragraph handler, and emit header for it
        handler = handler_cls(app=app, **mo.groupdict())

        yield from handler.header()

def _process_docstring(app, lines, transform, options):
    if transform != app.config.hawkmoth_javadoc_transform:
        return

    lines[:] = [line for line in _convert(app=app, lines=lines)]

def process_docstring(lines):
    """Simple interface for CLI and testing."""
    lines[:] = [line for line in _convert(lines=lines)]

def setup(app):
    app.setup_extension('hawkmoth')

    app.add_config_value('hawkmoth_javadoc_transform', 'javadoc', 'env', [str])

    app.connect('hawkmoth-process-docstring', _process_docstring)

    return {
        'parallel_read_safe': True,
    }



# Copyright (c) 2016-2017, Jani Nikula <jani@nikula.org>
# Licensed under the terms of BSD 2-Clause, see LICENSE for details.
"""
Alternative docstring syntax
============================

This module abstracts different compatibility options converting different
syntaxes into 'native' reST ones.
"""

import re
from sphinx.util import logging

logger = logging.getLogger(__name__)
logger.warning('hawkmoth.util.doccompat module has been deprecated.')

# Basic Javadoc/Doxygen/kernel-doc import
#
# FIXME: try to preserve whitespace better

def javadoc(comment):
    """Basic javadoc conversion to reStructuredText"""

    # @param
    comment = re.sub(r"(?m)^([ \t]*)@param([ \t]+)([a-zA-Z0-9_]+|\.\.\.)([ \t]+)",
                     "\n\\1:param\\2\\3:\\4", comment)
    # @param[direction]
    comment = re.sub(r"(?m)^([ \t]*)@param\[([^]]*)\]([ \t]+)([a-zA-Z0-9_]+|\.\.\.)([ \t]+)",
                     "\n\\1:param\\3\\4: *(\\2)* \\5", comment)
    # @return
    comment = re.sub(r"(?m)^([ \t]*)@returns?([ \t]+|$)",
                     "\n\\1:return:\\2", comment)
    # @code/@endcode blocks. Works if the code is indented.
    comment = re.sub(r"(?m)^([ \t]*)@code([ \t]+|$)",
                     "\n::\n", comment)
    comment = re.sub(r"(?m)^([ \t]*)@endcode([ \t]+|$)",
                     "\n", comment)
    # Ignore @brief.
    comment = re.sub(r"(?m)^([ \t]*)@brief[ \t]+", "\n\\1", comment)

    # Ignore groups
    comment = re.sub(r"(?m)^([ \t]*)@(defgroup|addtogroup)[ \t]+[a-zA-Z0-9_]+[ \t]*",
                     "\n\\1", comment)
    comment = re.sub(r"(?m)^([ \t]*)@(ingroup|{|}).*", "\n", comment)

    return comment

def javadoc_liberal(comment):
    """Liberal javadoc conversion to reStructuredText"""

    comment = javadoc(comment)

    # Liberal conversion of any @tags, will fail for @code etc. but don't
    # care.
    comment = re.sub(r"(?m)^([ \t]*)@([a-zA-Z0-9_]+)([ \t]+)",
                     "\n\\1:\\2:\\3", comment)

    return comment

def kerneldoc(comment):
    """Basic kernel-doc conversion to reStructuredText"""

    comment = re.sub(r"(?m)^([ \t]*)@(returns?|RETURNS?):([ \t]+|$)",
                     "\n\\1:return:\\3", comment)
    comment = re.sub(r"(?m)^([ \t]*)@([a-zA-Z0-9_]+|\.\.\.):([ \t]+)",
                     "\n\\1:param \\2:\\3", comment)

    return comment

def convert(comment, **options):
    """Convert documentation from a supported syntax into reST."""

    transform = options.get('transform')

    transformations = {
        'javadoc-basic': javadoc,
        'javadoc-liberal': javadoc_liberal,
        'kernel-doc': kerneldoc,
    }

    if transform in transformations:
        comment = transformations[transform](comment)

    return comment



# Copyright (c) 2021, Jani Nikula <jani@nikula.org>
# Licensed under the terms of BSD 2-Clause, see LICENSE for details.

def string_list(argument):
    if argument is None:
        return []

    return [s.strip() for s in argument.split(',') if len(s.strip()) > 0]



# Copyright (c) 2016-2023 Jani Nikula <jani@nikula.org>
# Copyright (c) 2018-2024 Bruno Santos <brunomanuelsantos@tecnico.ulisboa.pt>
# Licensed under the terms of BSD 2-Clause, see LICENSE for details.

from clang.cindex import (
    TokenKind,
    CursorKind,
    TypeKind,
    StorageClass,
    AccessSpecifier,
    ExceptionSpecificationKind,
    SourceLocation,
    SourceRange,
)


def _get_semantic_parent_namespace(cursor, namespace):
    semantic_parent = cursor.semantic_parent
    if not semantic_parent:
        return namespace

    if semantic_parent.kind == CursorKind.NAMESPACE:
        # parent is a namespace => add namespace in front
        if namespace is not None:
            namespace = f'{semantic_parent.spelling}::{namespace}'
        else:
            namespace = semantic_parent.spelling
        # check again for nested namespaces
        return _get_semantic_parent_namespace(semantic_parent, namespace)

    return namespace


class DocCursor:
    """Documentation centric wrapper for Clang's own ``Cursor``.

    This class abstracts a documentation worthy cursor so the user can query
    relevant bits for documentation purpose, but otherwise hide all the
    complications behind Clang's AST traversal and extraction of said bits of
    information.

    Technically, this class can hold any Clang cursor within itself, but it
    won't expose any relevant information for those.
    """

    def __init__(self, domain=None, cursor=None, comments=None):
        self._comments = comments if comments else {}
        self._cc = cursor
        self._domain = domain

        if self._cc.hash in self._comments:
            self._comment = self._comments[self._cc.hash]
        else:
            self._comment = None

    def __hash__(self):
        return self._cc.hash

    @property
    def meta(self):
        return {
            'line': self.line,
            'cursor.kind': self._cc.kind,
            'cursor.displayname': self._cc.displayname,
            'cursor.spelling': self._cc.spelling,
        }

    @property
    def location(self):
        return self._cc.location

    @property
    def comment(self):
        return self._comment.spelling if self._comment else None

    @property
    def domain(self):
        return self._domain

    @property
    def kind(self):
        return self._cc.kind

    @property
    def name(self):
        return self.namespace_prefix + self._cc.spelling if self._cc.spelling else self.decl_name

    @property
    def decl_name(self):
        if self._cc.kind in [CursorKind.VAR_DECL, CursorKind.FIELD_DECL]:
            return self._var_type_fixup(self)[1]
        if self._cc.kind in [CursorKind.STRUCT_DECL,
                             CursorKind.UNION_DECL,
                             CursorKind.ENUM_DECL,
                             CursorKind.CLASS_DECL,
                             CursorKind.CLASS_TEMPLATE,
                             CursorKind.TYPE_ALIAS_TEMPLATE_DECL]:
            return self._type_definition_fixup()
        else:
            # self.name would recurse back here if self._cc.spelling is None
            return self.namespace_prefix + self._cc.spelling if self._cc.spelling else None

    @property
    def namespace_prefix(self):
        if self.domain != 'cpp':
            return ''
        namespace = _get_semantic_parent_namespace(self._cc, None)
        return f'{namespace}::' if namespace else ''

    @property
    def type(self):
        if self._cc.kind in [CursorKind.VAR_DECL, CursorKind.FIELD_DECL]:
            return self._var_type_fixup(self)[0]
        if self._cc.kind == CursorKind.FUNCTION_DECL:
            return self._function_fixup()
        if self._cc.kind in [CursorKind.CONSTRUCTOR,
                             CursorKind.DESTRUCTOR,
                             CursorKind.CXX_METHOD,
                             CursorKind.FUNCTION_TEMPLATE]:
            return self._method_fixup()
        else:
            return self._cc.type.spelling

    @property
    def line(self):
        return self._comment.extent.start.line if self._comment else None

    @property
    def args(self):
        if self._cc.kind == CursorKind.MACRO_DEFINITION:
            return self._get_macro_args()
        if self._cc.kind in [CursorKind.FUNCTION_DECL,
                             CursorKind.CONSTRUCTOR,
                             CursorKind.DESTRUCTOR,
                             CursorKind.CXX_METHOD,
                             CursorKind.FUNCTION_TEMPLATE]:
            return self._get_fn_args()
        else:
            return None

    @property
    def quals(self):
        if self._cc.kind == CursorKind.FUNCTION_DECL:
            return ''
        if self._cc.kind in [CursorKind.CONSTRUCTOR,
                             CursorKind.DESTRUCTOR,
                             CursorKind.CXX_METHOD,
                             CursorKind.FUNCTION_TEMPLATE]:
            return ' '.join(self._get_method_quals()[1])
        else:
            return None

    @property
    def is_scoped_enum(self):
        return self._cc.is_scoped_enum()

    @property
    def value(self):
        if self._cc.kind == CursorKind.ENUM_CONSTANT_DECL:
            if '=' in [t.spelling for t in self.get_tokens()]:
                return self._cc.enum_value
            else:
                return None
        if self._cc.kind in [CursorKind.TYPE_ALIAS_DECL, CursorKind.TYPE_ALIAS_TEMPLATE_DECL]:
            return self._get_underlying_type()
        else:
            return None

    def get_children(self):
        """Get children cursors."""
        domain = self.domain

        # Identify `extern "C"` blocks and change domain accordingly.
        # Prior to Clang 18, the Python bindings don't return the cursor kind
        # LINKAGE_SPEC as one would expect, so we need to do it the hard way.
        if domain == 'cpp' and self.kind in [CursorKind.LINKAGE_SPEC,
                                             CursorKind.UNEXPOSED_DECL]:
            tokens = self.get_tokens()
            ntoken = next(tokens, None)
            if ntoken and ntoken.spelling == 'extern':
                ntoken = next(tokens, None)
                if ntoken and ntoken.spelling == '"C"':
                    domain = 'c'

        for c in self._cc.get_children():
            yield DocCursor(domain=domain, cursor=c, comments=self._comments)

    def get_tokens(self):
        """Get cursor tokens.

        Wrapper for Clang's `cursor.get_tokens()` that addresses issues for
        cursors whose extent contains macro expansions. The result may be empty
        or contain bogus tokens, depending on the case.

        The problem seems to be related to `cursor.extent`. Recreating the
        extent and getting the tokens from the translation unit works fine. The
        `__repr__` for both the recreated and original extents is the same, but
        comparison indicates they do differ under the hood.
        """
        tu = self._cc.translation_unit

        start = self._cc.extent.start
        start = SourceLocation.from_position(tu, start.file, start.line, start.column)

        end = self._cc.extent.end
        end = SourceLocation.from_position(tu, end.file, end.line, end.column)

        extent = SourceRange.from_locations(start, end)

        yield from tu.get_tokens(extent=extent)

    def _get_fn_args(self):
        """Get function / method arguments."""
        args = []

        # Only fully prototyped functions will have argument lists to process.
        if self._cc.type.kind == TypeKind.FUNCTIONPROTO:
            for c in self.get_children():
                if c._cc.kind == CursorKind.PARM_DECL:
                    arg_ttype, arg_name = self._var_type_fixup(c)
                    args.extend([(arg_ttype, arg_name)])

            if self._cc.type.is_function_variadic():
                args.extend([('', '...')])
            if len(args) == 0:
                args.extend([('', 'void')])

        return args

    def _function_fixup(self):
        """Parse additional details of a function declaration."""
        full_type = self._get_function_quals()

        template_line = self._get_template_line()
        if template_line:
            full_type.append(template_line)

        full_type.append(self._normalize_type(self._cc.result_type.spelling))

        ttype = ' '.join(full_type)

        return ttype

    def _method_fixup(self):
        """Parse additional details of a method declaration."""
        full_type = []

        access_spec = self._get_access_specifier()
        if access_spec:
            full_type.append(access_spec)

        pre_quals, _ = self._get_method_quals()

        full_type.extend(pre_quals)

        template_line = self._get_template_line()
        if template_line:
            full_type.append(template_line)

        if self._cc.kind not in [CursorKind.CONSTRUCTOR, CursorKind.DESTRUCTOR]:
            full_type.append(self._cc.result_type.spelling)

        ttype = ' '.join(full_type)

        return ttype

    def _type_definition_fixup(self):
        """Fix non trivial type definitions."""
        type_elem = []

        # Short cut for anonymous symbols.
        if self._cc.is_anonymous():
            return None

        # libclang 16 and later have cursor.spelling == cursor.type.spelling
        # for typedefs of anonymous entities, while libclang 15 and earlier
        # have an empty string. Match the behaviour across libclang versions.
        if self._cc.spelling == '':
            return self._cc.type.spelling

        type_elem.extend(self._specifiers_fixup(self._cc.type))

        colon_suffix = ''
        if self._cc.kind in [CursorKind.STRUCT_DECL,
                             CursorKind.CLASS_DECL,
                             CursorKind.CLASS_TEMPLATE]:
            inheritance = self._get_inheritance()
            if inheritance:
                colon_suffix = inheritance
        elif self._cc.kind == CursorKind.ENUM_DECL:
            scopedenum_type = self._get_scopedenum_type()
            if scopedenum_type:
                colon_suffix = scopedenum_type

        template = self._get_template_line()
        template = template + ' ' if template else ''

        return f'{template}{self.namespace_prefix}{self._cc.spelling}{colon_suffix}'

    def _get_macro_args(self):
        """Get macro arguments.

        Returns:
            None for simple macros, a potentially empty list of arguments for
            function-like macros
        """
        tokens = self.get_tokens()

        # Use the first two tokens to make sure this starts with 'IDENTIFIER('
        # *without* a space before the paren.
        identifier = next(tokens)
        paren = next(tokens, None)
        if paren is None or identifier.extent.end != paren.extent.start or paren.spelling != '(':
            return None

        # Naïve parsing of macro arguments
        args = []
        arg_spellings = []
        for token in tokens:
            if token.spelling in [')', ',']:
                if arg_spellings:
                    args.extend([('', ''.join(arg_spellings))])
                    arg_spellings = []

                if token.spelling == ')':
                    return args
                continue
            elif token.kind == TokenKind.IDENTIFIER or token.spelling == '...':
                arg_spellings.append(token.spelling)
            else:
                break

        return None

    def _symbolic_dims(self):
        dim = None
        for spelling in [t.spelling for t in self.get_tokens()]:
            if spelling == '[':
                # dim should be None here
                dim = []
            elif spelling == ']':
                # dim should not be None here
                yield ' '.join(dim)
                dim = None
            elif dim is not None:
                dim.append(spelling)

    def _dims_fixup(self, dims):
        if not dims:
            return ''

        symbolic_dims = list(self._symbolic_dims())
        if len(symbolic_dims) == len(dims):
            dims = symbolic_dims

        return ''.join([f'[{d}]' for d in dims])

    def _get_storage_class(self):
        """Get the storage class of a cursor.

        Only storage classes that are relevant to the documentation are
        returned.

        Returns:
            Storage class as a string. ``None`` otherwise.
        """
        storage_class_map = {
            StorageClass.EXTERN: 'extern',
            StorageClass.STATIC: 'static',
        }

        return storage_class_map.get(self._cc.storage_class)

    def _get_function_quals(self):
        """Get all the qualifiers of a function object.

        Returns:
            List of (prefix) function qualifiers.
        """
        tokens = [t.spelling for t in self.get_tokens()]
        quals = []

        if 'static' in tokens:
            quals.append('static')
        if 'inline' in tokens:
            quals.append('inline')

        return quals

    def _get_method_quals(self):
        """Get all the qualifiers of a method.

        Returns:
            List of prefix method qualifiers and list of suffix method
            qualifiers.
        """
        tokens = [t.spelling for t in self.get_tokens()]
        pre_quals = []
        pos_quals = []

        if self._cc.is_static_method():
            pre_quals.append('static')
        if self._cc.is_virtual_method():
            pre_quals.append('virtual')
        if 'constexpr' in tokens:
            pre_quals.append('constexpr')

        if self._cc.is_const_method():
            pos_quals.append('const')
        if self._cc.is_pure_virtual_method():
            pos_quals.append('= 0')
        if self._cc.is_default_method():
            pos_quals.append('= default')
        if 'delete' in tokens:
            pos_quals.append('= delete')
        if 'override' in tokens:
            pos_quals.append('override')

        except_spec = self._cc.exception_specification_kind
        if except_spec == ExceptionSpecificationKind.BASIC_NOEXCEPT:
            pos_quals.append('noexcept')

        return pre_quals, pos_quals

    def _specifiers_fixup(self, basetype):
        """Fix the type for C++ specifiers.

        Note the ``basetype`` is not necessarily ``cursor.type``. When dealing
        with pointers or arrays, we need to get to the base type as in
        :py:func:`_var_type_fixup`.

        Returns:
            List of C++ specifiers for the cursor.
        """
        tokens = [t.spelling for t in self.get_tokens()]
        type_elem = []

        if 'mutable' in tokens:
            type_elem.append('mutable')

        # If 'constexpr', strip the redundant 'const' that Clang adds to the
        # type spelling by default.
        if 'constexpr' in tokens:
            type_elem.append('constexpr')
            type_elem.append(basetype.spelling[len('const '):])
        else:
            type_elem.append(basetype.spelling)

        return type_elem

    def _get_access_specifier(self):
        """Get the access specifier of a cursor, if any.

        Returns:
            One of 'private', 'protected', 'public' or `None`.
        """
        # No access specifiers in C.
        if self.domain == 'c':
            return None

        # No access specifiers in redundant contexts.
        if self._cc.semantic_parent:
            if self._cc.semantic_parent.kind == CursorKind.UNION_DECL:
                return None

        name_map = {
            AccessSpecifier.PRIVATE: 'private',
            AccessSpecifier.PROTECTED: 'protected',
            AccessSpecifier.PUBLIC: 'public',
        }

        return name_map.get(self._cc.access_specifier, None)

    def _get_template_line(self):
        """Get the template arguments of a cursor.

        This recurses for templated template arguments.

        Returns:
            String with the form 'template<...> ' if the cursor is a template
            or `None` otherwise. When the cursor represents a templated
            template argument, the returned string is actually of the form
            'template<...> name', but this should only occur under recursion.
        """
        # We only add the name when we recurse, in which case we need to track
        # the name of the templated template argument. Otherwise the name is
        # not part of the template arguments.
        name = ''

        if self._cc.kind not in [CursorKind.CLASS_TEMPLATE,
                                 CursorKind.FUNCTION_TEMPLATE,
                                 CursorKind.TEMPLATE_TEMPLATE_PARAMETER,
                                 CursorKind.TYPE_ALIAS_TEMPLATE_DECL]:
            return None

        # The type of type parameters can be 'typename' and 'class'. These are
        # equivalent, but we want it to look like the source code for
        # consistency. We can do it by looking at the tokens directly. This is
        # slightly complicated due to variadic template type parameters.
        def typetype(cursor):
            tokens = list(cursor.get_tokens())
            if tokens[-2].spelling == '...':
                return f'{tokens[-3].spelling}...'
            else:
                return f'{tokens[-2].spelling}'

        # We need to add the keyword 'typename' or 'class' if we have recursed
        # and therefore we are inside the template argument list.
        if self._cc.kind == CursorKind.TEMPLATE_TEMPLATE_PARAMETER:
            name = f' {typetype(self)} {self._cc.spelling}'

        template_args = []
        for child in self.get_children():
            if child._cc.kind == CursorKind.TEMPLATE_TYPE_PARAMETER:
                template_args.append(f'{typetype(child)} {child._cc.spelling}')
            elif child._cc.kind == CursorKind.TEMPLATE_NON_TYPE_PARAMETER:
                arg_name = f' {child._cc.spelling}' if child._cc.spelling != '' else '...'
                template_args.append(f'{child._cc.type.spelling}{arg_name}')
            elif child._cc.kind == CursorKind.TEMPLATE_TEMPLATE_PARAMETER:
                arg = child._get_template_line()
                if arg:
                    template_args.append(arg)

        return f'template<{", ".join(template_args)}>{name}'

    def _get_inheritance(self):
        """Get the full inheritance list of a cursor in C++ syntax.

        Returns:
            String with the form ': A, B, ...' when a cursor has
            `CXX_BASE_SPECIFIER` children or `None` otherwise.
        """
        inherited = []
        for child in self.get_children():
            if child._cc.kind == CursorKind.CXX_BASE_SPECIFIER:
                def pad(s): return s + ' ' if s else ''
                access_spec = child._get_access_specifier()
                if child._cc.referenced.kind == CursorKind.CLASS_DECL:
                    # use referenced type if possible for full namespace
                    spelling = child._cc.referenced.type.spelling
                else:
                    spelling = child._cc.type.spelling
                inherited.append(f'{pad(access_spec)}{spelling}')

        return ': ' + ', '.join(inherited) if len(inherited) > 0 else None

    def _get_scopedenum_type(self):
        """Get the explicit underlying type of a scoped enumerator.

        Returns:
            Underlying type of a scoped enumerator that has been explicitly
            defined. ``None`` otherwise.
        """
        if self._cc.kind == CursorKind.ENUM_DECL and self._cc.is_scoped_enum():
            if list(self.get_tokens())[3].spelling == ':':
                return f': {self._cc.enum_type.spelling}'
        return None

    @staticmethod
    def _normalize_type(type_string):
        return 'bool' if type_string == '_Bool' else type_string

    @staticmethod
    def _var_type_fixup(cursor):
        """Fix non trivial variable and argument types.

        If this is an array, the dimensions should be applied to the name, not
        the type. If this is a function pointer, or an array of function
        pointers, the name should be within the parenthesis as in ``(*name)``
        or ``(*name[N])``.
        """
        cursor_type = cursor._cc.type

        stars_and_quals = ''
        dims = []
        while True:
            if cursor_type.kind == TypeKind.POINTER:
                quals = []
                if cursor_type.is_const_qualified():
                    quals.append('const')
                if cursor_type.is_volatile_qualified():
                    quals.append('volatile')
                if cursor_type.is_restrict_qualified():
                    quals.append('restrict')

                spacer = ' ' if quals and stars_and_quals else ''
                stars_and_quals = '*' + ' '.join(quals) + spacer + stars_and_quals

                cursor_type = cursor_type.get_pointee()
            elif cursor_type.kind == TypeKind.CONSTANTARRAY:
                dims.append(cursor_type.element_count)
                cursor_type = cursor_type.get_array_element_type()
            elif cursor_type.kind == TypeKind.INCOMPLETEARRAY:
                dims.append('')
                cursor_type = cursor_type.get_array_element_type()
            else:
                break

        dims = cursor._dims_fixup(dims)

        type_elem = []

        access_spec = cursor._get_access_specifier()
        if access_spec:
            type_elem.append(access_spec)

        if cursor_type.kind == TypeKind.FUNCTIONPROTO:
            def pad(s): return s if s.endswith('*') or s.endswith('&') else s + ' '

            args = []
            for c in cursor.get_children():
                if c._cc.kind == CursorKind.PARM_DECL:
                    arg_ttype, arg_name = cursor._var_type_fixup(c)
                    args.append(f'{pad(arg_ttype)}{arg_name}' if arg_name else arg_ttype)
            if cursor_type.is_function_variadic():
                args.append('...')
            if len(args) == 0:
                args.append('void')

            ret_type = cursor._normalize_type(cursor_type.get_result().spelling)

            name = f'''{pad(ret_type)}({pad(stars_and_quals)}{cursor._cc.spelling}{dims})({', '.join(args)})'''  # noqa: E501
        else:

            storage_class = cursor._get_storage_class()
            if storage_class:
                type_elem.append(storage_class)

            type_elem.extend(cursor._specifiers_fixup(cursor_type))

            if stars_and_quals:
                type_elem.append(stars_and_quals)

            name = cursor._cc.spelling + dims

        # Convert _Bool to bool
        type_elem = [cursor._normalize_type(t) for t in type_elem]

        ttype = ' '.join(type_elem)
        return ttype, name

    def _get_underlying_type(self):
        if self._cc.kind == CursorKind.TYPE_ALIAS_DECL:
            return self._cc.underlying_typedef_type
        elif self._cc.kind == CursorKind.TYPE_ALIAS_TEMPLATE_DECL:
            for child in self._cc.get_children():
                if child.kind == CursorKind.TYPE_ALIAS_DECL:
                    return child.underlying_typedef_type
        return None



# Copyright (c) 2021 Jani Nikula <jani@nikula.org>
# Licensed under the terms of BSD 2-Clause, see LICENSE for details.
"""
Compiler helpers
================

This module provides helper functions to access compiler information outside of
the Clang Python Bindings, for example system include paths.
"""

import subprocess
from sphinx.util import logging

logger = logging.getLogger(__name__)

def _removesuffix(s, suffix):
    if suffix and s.endswith(suffix):
        return s[:-len(suffix)]
    else:
        return s[:]

def _get_paths_from_output(output):
    started = False
    for line in output.splitlines():
        if not started:
            if line == '#include <...> search starts here:':
                started = True
            continue

        if line == 'End of search list.':
            break

        # Clang on macOS may print this.
        line = _removesuffix(line, '(framework directory)')

        yield line.strip()

def _get_include_paths(cpath, lang):
    try:
        result = subprocess.run([cpath, '-x', lang, '-E', '-Wp,-v', '-'],
                                stdin=subprocess.DEVNULL,
                                capture_output=True,
                                check=True,
                                text=True)
    except FileNotFoundError:
        logger.warning(f"get_include_args: {lang} compiler not found ('{cpath}')")
        return []

    except subprocess.CalledProcessError:
        logger.warning(f"get_include_args: incompatible {lang} compiler ('{cpath}')")
        return []

    if result.returncode != 0:
        logger.warning(f"get_include_args: incompatible {lang} compiler ('{cpath}')")
        return []

    return _get_paths_from_output(result.stderr)

def get_include_args(cpath='clang', lang='c', cc_path=None):
    if cc_path is not None:
        cpath = cc_path
        logger.warning('get_include_args: `cc_path` argument has been deprecated; use `cpath` instead')  # noqa: E501

    return ['-nostdinc'] + [f'-isystem{path}' for path in _get_include_paths(cpath, lang)]

if __name__ == '__main__':
    import argparse
    import pprint

    parser = argparse.ArgumentParser()
    parser.add_argument('compiler', nargs='?', default='clang')
    parser.add_argument('--lang', choices=['c', 'c++'], default='c')

    args = parser.parse_args()

    pprint.pprint(get_include_args(cpath=args.compiler, lang=args.lang))
